%% lab4 - Executes lab4 simulation according to the given config-id
%
% Parameters:
%    testId  -- simulation ID
%
%  Filename: lab4.m
%  Revision: 0.2
%  Date:     2012-05-15
%  Author:   Mikica B Kocic

function lab4( varargin )

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %% Initialize & do the requested simulation.

    [ params, body ] = lab4_defs( varargin{:} );  % Setup simulation test-bed
   
    result = WoRB( params, body );                % Call WoRB mex to perform simulation

    if isempty( result )
        params.ConservationPlots = false;         % Disable diagrams if no result re
    else
        result( isnan(result(:,1)), : ) = [];     % Remove time-steps without data
    end

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %% Post processing: plot energies, momentum and interactions.

    % Determine total number of subplots

    if size( result, 1 ) <= 1 || ~params.ConservationPlots
        subplot_count = 0;
    else
        subplot_count = params.PlotContacts + params.PlotEnergy + params.PlotHeight ...
                      + params.PlotLinearMomentum + params.PlotAngularMomentum;
    end

    if subplot_count

        % Setup the figure name (windows title)
        fig_name = '';
        if params.PlotContacts
            fig_name = [ fig_name, '/Collisions' ];
        end
        if params.PlotEnergy
            fig_name = [ fig_name, '/Energies' ];
        end
        if params.PlotLinearMomentum
            fig_name = [ fig_name, '/Linear Momentum' ];
        end
        if params.PlotAngularMomentum
            fig_name = [ fig_name, '/Angular Momentum' ];
        end
        % Get rid of leading '/' and add the simulation ID as the prefix
        fig_name = fig_name(2:end);
        fig_name = [ params.Title, ' (', params.ID, ')', ' - ', fig_name,  ]; 

        % Get the size of the screen
        screen.Rect = get( 0, 'ScreenSize' ); % gets [ left, bottom, width, height ]
        screen.L = screen.Rect(1);  screen.B = screen.Rect(2);
        screen.W = screen.Rect(3);  screen.H = screen.Rect(4);

        % Calculate position (*Warning*: Reuse screen structure of the main figure!)
        screen.fig_W = screen.H * 2/3; % Width  = 2/3 of the screen height (not width)
        screen.fig_H = screen.H * 2/3; % Height = 2/3 of the screen height
        screen.fig_L = ( screen.W - screen.fig_W ) * 0.3 + 30; % Figure left position
        screen.fig_B = ( screen.H - screen.fig_H ) * 0.6 - 30; % Figure bottom position

        % Reduce height depending of the number of the subplots
        max_plots = max( 3, subplot_count );
        plot_offset = screen.fig_H * ( max_plots - subplot_count ) / max_plots;
        screen.fig_H = screen.fig_H - plot_offset;
        screen.fig_B = screen.fig_B + plot_offset;

        % Setup paper size
        screen.pap_W = params.PlotWidth;
        screen.pap_H = params.SubplotHeight * subplot_count;

        % Create figure
        energyFig = figure( 'Name', fig_name, 'NumberTitle', 'off', ...
            'FileName', [ 'lab4_fig2_', params.ID, '.fig' ], ... % name of the FIG-file 
            'Position', [ screen.fig_L, screen.fig_B, screen.fig_W, screen.fig_H ], ...
            'PaperPositionMode', 'manual', 'PaperUnits', 'centimeters', ... 
            'PaperSize', [ screen.pap_W, screen.pap_H ], ...
            'PaperPosition', [ 0, 0, screen.pap_W, screen.pap_H ] ...
        );

        % Allocate space for individual plots axes handles
        ax = zeros( subplot_count, 1 );
        if subplot_count == 1
            ax(1) = gca;
        end
        cur_subplot = 0;  % Track the current subplot number
    end

    %%  Plot collision statistics per time-step

    if params.PlotContacts && subplot_count
        drawnow; % Flush event queue and update figure window
        set( 0, 'CurrentFigure', energyFig ); % Go back to energyFig

        % Create as a subplot, if more than one plot
        cur_subplot = cur_subplot + 1;
        if subplot_count > 1
            ax( cur_subplot ) = subplot( subplot_count, 1, cur_subplot );
        end
        hold on; % Retain subsequent plots in figure

        % Title, axes and grid:
        title( 'Collision Statistics', 'FontSize', 11 );
        set( ax( cur_subplot ) , 'YGrid', 'on' );
        ylabel( { 'Total # of contacts'; 'per time-step' }, ...
            'FontSize', 11, 'FontName', 'Times', 'Margin', 1 );

        % Plot statistics for total number of collisions
        i = find( result(:,2) > 0 );
        if ~isempty(i)
            stem( result(i,1), result(i,2), 'DisplayName', 'contact', ...
                'Color', [ 0.4 0.8 0.8 ], 'Marker', '.', ...
                'MarkerFaceColor', 'k', 'MarkerEdgeColor', 'none' );
        end

        % Finally, we have plots so we can place legend
        set( legend( 'show' ), 'Location', 'SouthEast', 'FontSize', 9 );
    end

    %%  Plot the kinetic, potential and total energy of the system over time

    if params.PlotEnergy && subplot_count
        drawnow; % Flush event queue and update figure window
        set( 0, 'CurrentFigure', energyFig ); % Go back to energyFig

        % Create as a subplot, if more than one plot
        cur_subplot = cur_subplot + 1;
        if subplot_count > 1
            ax( cur_subplot ) = subplot( subplot_count, 1, cur_subplot );
        end
        hold on; % Retain subsequent plots in figure

        % Title, axes and grid:
        title( 'Energy', 'FontSize', 11 );
        grid on;
        ylabel( '{\itE} /\fontsize{11}{\rmJ}', ...
            'FontSize', 12, 'FontName', 'Times' );

        % Plots:
        E_tot = result(:,3) + result(:,4); % = E_k + E_p
        plot( result(:,1), result(:,3), 'b-', 'LineWidth', 0.5 ); % E_k, blue
        plot( result(:,1), result(:,4), 'r-', 'LineWidth', 0.5 ); % E_p, red
        plot( result(:,1), E_tot,       'k-', 'LineWidth', 1.5 ); % E_tot, black, thick

        % Legend:
        set( legend( '{\itE}_{\rmk}', '{\itE}_{\rmp}', '{\itE}_{\rmtot}' ), ...
            'FontSize', 10, 'FontName', 'Times', 'Location', 'SouthEast' );
    end

    %%  Plot the linear momentum of the system over time

    if params.PlotLinearMomentum && subplot_count
        drawnow; % Flush event queue and update figure window
        set( 0, 'CurrentFigure', energyFig ); % Go back to energyFig

        % Create as a subplot, if more than one plot
        cur_subplot = cur_subplot + 1;
        if subplot_count > 1
            ax( cur_subplot ) = subplot( subplot_count, 1, cur_subplot );
        end
        hold on; % Retain subsequent plots in figure

        % Title, axes and grid:
        title( 'Linear Momentum', 'FontSize', 11 );
        grid on;
        ylabel( '{\itp} /\fontsize{11}({\rmkg \cdot m/s})', ...
            'FontSize', 12, 'FontName', 'Times' );

        ctype = { 'x',  'y',  'z'  };
        ltype = { 'r-', 'g-', 'b-' };

        % Plot linear-momentum components:
        for i = 1:3
            plot( result(:,1), result(:,4+i), ltype{i}, 'DisplayName', sprintf( ...
                '\\fontname{Times New Roman}\\fontsize{11}{\\itp}_{\\it%s\\rm,tot}', ...
                ctype{i} ...
            ));
        end
        set( legend( 'show' ), 'Location', 'SouthEast' );
    end

    %%  Plot the angular momentum 

    if params.PlotAngularMomentum && subplot_count
        drawnow; % Flush event queue and update figure window
        set( 0, 'CurrentFigure', energyFig ); % Go back to energyFig

        % Create as a subplot, if more than one plot
        cur_subplot = cur_subplot + 1;
        if subplot_count > 1
            ax( cur_subplot ) = subplot( subplot_count, 1, cur_subplot );
        end
        hold on; % Retain subsequent plots in figure

        % Title, axes and grid:
        title( 'Angular Momentum', 'FontSize', 11 );
        grid on;
        ylabel( '{\itL} /\fontsize{11}({\rmkg \cdot m^2/s})', ...
            'FontSize', 12, 'FontName', 'Times' );

        % Configure plot colors and legend
        ctype = { 'x',  'y',  'z'  };
        ltype = { 'r-', 'g-', 'b-' };

        % Plot angular velocity components for the first particle:
        for i = 1:3
            plot( result(:,1), result(:,7+i), ltype{i}, 'DisplayName', sprintf( ...
                '\\fontname{Times New Roman}\\fontsize{11}{\\itL}_{\\it%s}', ...
                ctype{i} ...
            ));
        end
        set( legend( 'show' ), 'Location', 'SouthEast' );
    end

    %%  Plot the height of the barycenter of the system over time

    if params.PlotHeight && subplot_count
        drawnow; % Flush event queue and update figure window
        set( 0, 'CurrentFigure', energyFig ); % Go back to energyFig

        % Create as a subplot, if more than one plot
        cur_subplot = cur_subplot + 1;
        if subplot_count > 1
            ax( cur_subplot ) = subplot( subplot_count, 1, cur_subplot );
        end

        hold on; % Retain subsequent plots in figure

        % Title, axes and grid:
        title( 'Barycenter Height', 'FontSize', 11 );
        grid on;
        ylabel( '{\ity}_{\rmcm} /\fontsize{11}{\rmm}', ...
            'FontSize', 12, 'FontName', 'Times' );

        % Plots:
        plot( result(:,1), result(:,11), 'b-', 'LineWidth', 0.5 ); % height: blue, thick
        set( legend( 'height' ), 'Location', 'SouthEast', 'FontSize', 9 );
    end

    %%  Link axes and optionally save figure as EPS file

    if subplot_count

        % Label horizontal axis as time and link axes over time
        xlabel( '{\itt} /{\rms}', 'FontSize', 12, 'FontName', 'Times' );
        if subplot_count > 1
            linkaxes( flipdim( ax, 1 ), 'x' );
        end

        % Add context menu to export figure
        cxMenu = uicontextmenu( 'Parent', energyFig );
        uimenu( cxMenu, 'Label', 'Export as EPS', 'Callback', [ ...
            'exportFigure( gcbf, ''-depsc2'', ', ...
            '''lab4_fig2_', params.ID, ''' );' ...
        ]);
        uimenu( cxMenu, 'Label', 'Export as PNG', 'Callback', [ ...
            'exportFigure( gcbf, ''-dpng'', ', ...
            '''lab4_fig2_', params.ID, ''' );' ...
        ]);
        uimenu( cxMenu, 'Label', 'Export as EMF', 'Callback', [ ...
            'exportFigure( gcbf, ''-dmeta'', ', ...
            '''lab4_fig2_', params.ID, ''' );' ...
        ]);
        set( energyFig, 'UIContextMenu', cxMenu );
        set( ax, 'UIContextMenu', cxMenu );

        % Convert all 'Helvetica' fonts to 'Arial' and 'Times' to 'Times New Roman'
        set( findall( findall( energyFig, '-property', 'FontName' ), ...
            'FontName', 'Helvetica' ), 'FontName', 'Arial' );
        set( findall( findall( energyFig, '-property', 'FontName' ), ...
            'FontName', 'Times' ), 'FontName', 'Times New Roman' );

        drawnow; % Flush event queue and update figure window
        set( 0, 'CurrentFigure', energyFig ); % Go back to energyFig
    end

end % lab4 function
