%% Lab4: World of Rigid Bodies; Simulation test-bed constructor
% Constructs the default 'params' and 'body' structures which contains properties 
% that define the mathematical model and simulation constants, e.g., gravity, 
% friction, mass and geometry of the objects, visualization flags etc.
%
%  Filename: lab4_defs.m
%  Revision: 0.3
%  Date:     2012-05-15
%  Author:   Mikica B Kocic 
%
%  Usage:
%    [ params, body ] = lab4_defs;
%    [ params, body ] = lab4_defs( 'configId' );
%
%  Arguments:
%    configId: configuration profile ID
%

function [ params, body ] = lab4_defs( varargin )

    if nargin >= 1
        configId = varargin{1};
    else
        configId = 'Default';
    end

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %% Default parameters

    params = struct ();
    body   = struct ();

    params.Title = [ 'Lab4: ', configId ];
    params.ID    = datestr( now, 'YYYYmmdd_HHMMSS' );
    
    params.IsRunning            = true;   % Main-loop ends, when set to false
    params.IsPaused             = false;  % Is simulation paused: no
    params.AutoPause            = false;  % Is single-step mode: no
    params.Wireframe            = false;  % Show bodies in wireframe instead of solid: no
    params.ShowBodyAxes         = true;   % Show body axes: yes
    params.ShowFloorMirror      = false;  % Show objects mirrored in the floor: no
    params.ShowContacts         = false;  % Show contact normals: no
    params.ShowTrajectories     = false;  % Show object's trajectories: no
    params.ShowStateVariables   = true;   % Show system state variables: yes
    params.ShowHelp             = true;   % Show short help: yes
    params.DumpInitialState     = false;  % Dump initial parameters and bodies

    params.GridTickLength       = 1.0;    % Tick size of the grid, in meters
    params.GridTicks            = 50;     % Number of ticks in the grid

    params.FollowObject         = -1;     % Index of object to follow with camera (0 = #1)
    params.CameraAngle          = 55.0;   % Angle in degrees, left/right of x-axis
    params.CameraElevation      = 25.0;   % Angle in degres, up/bellow the horizon
    params.CameraZoom           = 15.0;   % Position in m, from the coordinate system origin
    params.CameraLookAt         = [ 0 0 0 ];  % Look at position

    params.TimeStep             = 0.01;   % Integrator time-step, in seconds
    params.TimeStepsPerFrame    = 1;      % Number of time-steps solved per one video frame
    params.TimeStepsPerSnapshot = 30;     % Number of time-steps per one trajectory snapshot
    params.FinalTime            = 0;      % Final time, in seconds (0 for infinite)

    params.Restitution          = 1.0;    % Coefficient of restitution, 0 = max damping
    params.Relaxation           = 0.2;    % Position projections relaxation, 0 = no relaxation
    params.Friction             = 0.0;    % Dynamic friction coefficient, 0 = no friciton
    params.Gravity              = [ 0 0 0 ]; % Acceleration of gravity, m s^-2

    params.ConservationPlots    = true;   % Plot energy etc graphs
    params.PlotContacts         = true;   % Plot number of collisions over time
    params.PlotEnergy           = true;   % Plot energy levels over time
    params.PlotLinearMomentum   = true;   % Plot linear momentum over time
    params.PlotAngularMomentum  = true;   % Plot angular momentum over time
    params.PlotHeight           = false;  % Plot barycenter height for the last body
    params.PlotWidth            = 17;     % Width of the plot, in centimeters
    params.SubplotHeight        = 5;      % Height of the subplot (when stacked), in cm

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %% Customize parameters and setup rigid bodies

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    configId = lower( configId );
    switch configId

    case { '1a', '1b', '1c' } 
        %% Two colliding plates (basic level task)

        params.FinalTime = 10;
        params.ShowTrajectories = true;
        params.FollowObject = -1;
        params.CameraLookAt = [ -2, 3, 0 ];  % Look at position
        
        V     = 1;                  % Initial velocity, m/s
        M     = 0.1;                % The default body mass, kg
        L     = 10;                 % Characteristic length of the system, m
        thick = 0.01;               % Box (plate if ~= 0) thickness, m
        box   = [ L, thick, L/2 ];  % Box or plate dimensions, m

        body(1).Geometry = 'cuboid';        % Geometry: 'cuboid' or 'sphere'
        body(1).HalfExtent = box / 2;       % Dimensions, m
        body(1).M = M;                      % Mass, kg
        body(1).X = [ -L/4, 3, 0 ];         % Position, m
        body(1).Q = [ cos(pi/4), 0, sin(pi/4), 0 ]; % Orientation: 90 around Y-axis
        body(1).V = [ 0, 0, 0 ];            % Velocity, m/s
        body(1).W = [ 0, 0, 0 ];            % Angular velocity, rad/s

        body(2).Geometry = 'cuboid';        % Geometry: 'cuboid' or 'sphere'
        body(2).HalfExtent = box / 2;       % Dimensions, m
        body(2).M = M;                      % Mass, kg
        body(2).X = [ (L/2 + V), 3, L/4 ];  % Position, m
        body(2).Q = [ 0, 0, 1, 0 ];         % Orientation: 180 around Y-axis
        body(2).V = [ -V, 0, 0 ];           % Velocity, m/s
        body(2).W = [ 0, 0, 0 ];            % Angular velocity, rad/s
            
        switch configId
            case '1b', body(2).Q(1) = body(2).Q(1) + 1e-4;
            case '1c', body(2).Q(1) = body(2).Q(1) - 1e-4;
        end

    case { 'default', '2a', '2b', '2c', '2d' } 
        %% Two colliding blocks thick 1 m having mass 10e3 kg
            
        params.TimeStepsPerSnapshot = 10;  % Number of time-steps per one trajectory snapshot
        
        if ~isequal( configId, 'default' )
            params.IsPaused = true;
            params.FinalTime = 15;
            params.ShowTrajectories = true;
            params.FollowObject = -1;
        else
            params.FinalTime = 60;
        end
        
        V     = 10;                 % Initial velocity, m/s
        M     = 10e3;               % The default body mass, kg
        L     = 10;                 % Characteristic length of the system, m
        thick = 1;                  % Box (plate if ~= 0) thickness, m
        box   = [ L, thick, L/2 ];  % Box or plate dimensions, m

        body(1).Geometry = 'cuboid';        % Geometry: 'cuboid' or 'sphere'
        body(1).HalfExtent = box / 2;       % Dimensions, m
        body(1).M = M;                      % Mass, kg
        body(1).X = [ -L/4, 3, 0 ];         % Position, m
        body(1).Q = [ cos(pi/4), 0, sin(pi/4), 0 ]; % Orientation: 90 around Y-axis
        body(1).V = [ 0, 0, 0 ];            % Velocity, m/s
        body(1).W = [ 0, 0, 0 ];            % Angular velocity, rad/s

        body(2).Geometry = 'cuboid';        % Geometry: 'cuboid' or 'sphere'
        body(2).HalfExtent = box / 2;       % Dimensions, m
        body(2).M = M;                      % Mass, kg
        body(2).X = [ (L/2 + V), 3, L/4 ];  % Position, m
        body(2).Q = [ 0, 0, 1, 0 ];         % Orientation: 180 around Y-axis
        body(2).V = [ -V, 0, 0 ];           % Velocity, m/s
        body(2).W = [ 0, 0, 0 ];            % Angular velocity, rad/s
        
        switch configId
            case '2b'
                body(2).Q(1) = body(2).Q(1) + 1e-4; % distort orientation
            case '2c'
                body(2).Q(1) = body(2).Q(1) - 1e-4; % distort orientation
            case { 'default', '2d' }
                body(2).Q(1) = body(2).Q(1) + 1e-4; % distort orientation
                body(2).X(2) = body(2).X(2) + 0.1;  % increase body 2 height
                params.FollowObject = 1;
                params.ShowFloorMirror = true;    % Show objects mirrored in floor
                if isequal( configId, 'default' )
                    params.ConservationPlots = false; % disable plots
                end
        end

    case '2e' 
        %% Two colliding blocks, one with mass 10^3 kg and the other 10^9 kg
            
        params.TimeStepsPerSnapshot = 10;  % Number of time-steps per one trajectory snapshot
        
        params.IsPaused = true;
        params.FinalTime = 15;
        params.ShowTrajectories = true;
        params.FollowObject = -1;
        
        V     = 10;                 % Initial velocity, m/s
        M     = 1e3;                % The default body mass, kg
        L     = 10;                 % Characteristic length of the system, m
        thick = 1;                  % Box (plate if ~= 0) thickness, m
        box   = [ L, thick, L/2 ];  % Box or plate dimensions, m

        body(1).Geometry = 'cuboid';        % Geometry: 'cuboid' or 'sphere'
        body(1).HalfExtent = box / 2;       % Dimensions, m
        body(1).M = 1e9;                    % Mass, kg
        body(1).X = [ -L/4, 3, 0 ];         % Position, m
        body(1).Q = [ cos(pi/4), 0, sin(pi/4), 0 ]; % Orientation: 90 around Y-axis
        body(1).V = [ 0, 0, 0 ];            % Velocity, m/s
        body(1).W = [ 0, 0, 0 ];            % Angular velocity, rad/s

        body(2).Geometry = 'cuboid';        % Geometry: 'cuboid' or 'sphere'
        body(2).HalfExtent = box / 2;       % Dimensions, m
        body(2).M = M;                      % Mass, kg
        body(2).X = [ (L/2 + V), 3, L/4 ];  % Position, m
        body(2).Q = [ 0, 0, 1, 0 ];         % Orientation: 180 around Y-axis
        body(2).V = [ -V, 0, 0 ];           % Velocity, m/s
        body(2).W = [ 0, 0, 0 ];            % Angular velocity, rad/s
        
        body(2).Q(1) = body(2).Q(1) + 1e-4; % distort body 2's orientation

    case { '3a', '3b', '3c', '3d' } 
        %% Two colliding jumping blocks; without and with dissipation
            
        params.TimeStepsPerSnapshot = 10;  % Number of time-steps per one trajectory snapshot
        
        params.IsPaused = true;
        params.FinalTime = 5;
        params.ShowTrajectories = true;
        params.PlotContacts = false;
        params.PlotLinearMomentum = false;
        params.PlotAngularMomentum = false;
        params.FollowObject = -1;
        params.Gravity = [ 0, -9.80665, 0 ];
        
        switch configId
            case '3a',  params.Restitution = 1.0;  params.Friction = 0.0; 
            case '3b',  params.Restitution = 0.6;  params.Friction = 0.0; 
            case '3c',  params.Restitution = 1.0;  params.Friction = 0.3; 
            case '3d',  params.Restitution = 0.6;  params.Friction = 0.3; 
        end
        
        V     = 10;                 % Initial velocity, m/s
        M     = 10e3;               % The default body mass, kg
        L     = 10;                 % Characteristic length of the system, m
        thick = 1;                  % Box (plate if ~= 0) thickness, m
        box   = [ L, thick, L/2 ];  % Box or plate dimensions, m

        body(1).Geometry = 'cuboid';        % Geometry: 'cuboid' or 'sphere'
        body(1).HalfExtent = box / 2;       % Dimensions, m
        body(1).M = M;                      % Mass, kg
        body(1).X = [ -L/4, 3, 0 ];         % Position, m
        body(1).Q = [ cos(pi/4), 0, sin(pi/4), 0 ]; % Orientation: 90 around Y-axis
        body(1).V = [ 0, 0, 0 ];            % Velocity, m/s
        body(1).W = [ 0, 0, 0 ];            % Angular velocity, rad/s

        body(2).Geometry = 'cuboid';        % Geometry: 'cuboid' or 'sphere'
        body(2).HalfExtent = box / 2;       % Dimensions, m
        body(2).M = M;                      % Mass, kg
        body(2).X = [ (L/2 + V), 3.5, L/4 ]; % Position, m
        body(2).Q = [ 1e-4, 0, 1, 0 ];      % Orientation: 180 around Y-axis
        body(2).V = [ -V, 0, 0 ];           % Velocity, m/s
        body(2).W = [ 0, 0, 0 ];            % Angular velocity, rad/s
        
    case '4a' 
        %% Cuboids falling onto the ground
            
        params.TimeStepsPerSnapshot = 10;  % Number of time-steps per one trajectory snapshot
        
        params.IsPaused = true;
        params.FinalTime = 0;
        params.ShowTrajectories = true;
        params.ShowBodyAxes = false;
        params.ConservationPlots = false;
        params.FollowObject = -1;
        params.Gravity = [ 0, -9.80665, 0 ];
        
        params.Restitution = 0.2;
        params.Friction = 0.2; 
        
        for i = 1:50
            body(i).Geometry = 'cuboid';         % Geometry: 'cuboid' or 'sphere'
            body(i).HalfExtent = 0.5 + ([1,2,3]-0.5).*rand(1,3); % Dimensions, m
            body(i).M = 1e3;                     % Mass, kg
            body(i).X = [ 0, (i-1)*1+0.5, 0 ];   % Position, m
            body(i).Q = rand(1,4);               % Orientation
            body(i).V = zeros(1,3);              % Velocity, m/s
            body(i).W = zeros(1,3);              % Angular velocity, rad/s
            body(i).ActiveColor = [ rand(1,3), 0.8 ]; % Random object colors
            body(i).CanBeDeactivated = true;     % Can the object be put a sleep
        end
        
    case '4b' 
        %% Cuboid falling onto stairs
            
        params.TimeStepsPerSnapshot = 10;  % Number of time-steps per one trajectory snapshot
        
        params.IsPaused = false;             % Not paused (starts immediately)
        params.FinalTime = 0;                % Infinite final time
        params.ShowTrajectories = true;      % Show trajectories
        params.ShowBodyAxes = false;         % Don't show body axes
        params.ShowStateVariables = false;   % Don't display state variables
        params.ShowHelp = false;             % Hide help
        params.ConservationPlots = false;    % Don't plot any data
        params.CameraAngle = 45;             % Setup camera angle
        params.CameraElevation = 10;         % Camera elevation = above horizon
        params.CameraLookAt = [ 0, 10, 10 ];  % Look at height 10 m
        params.CameraZoom = 20;              % Zoom out
        params.FollowObject = -1;            % Don't follow any object
        params.Gravity = [ 0, -9.80665, 0 ]; % Simulate with present gravity
        
        params.Restitution = 0.8;            % Coefficient of restitution
        params.Friction = 0.3;               % Coffficient of dynamic friction

        body(1).Geometry = 'cuboid';         % Geometry: 'cuboid' or 'sphere'
        body(1).HalfExtent = 0.15 + ([0.6,0.4,0.3]-0.15).*rand(1,3); % Dimensions, m
        body(1).M = 10;                      % Mass, kg
        body(1).X = [ 0, 15, -8.5 ];         % Position, m
        body(1).Q = rand(1,4);               % Orientation
        body(1).V = zeros(1,3);              % Velocity, m/s
        body(1).W = rand(1,3);               % Angular velocity, rad/s
        body(1).CanBeDeactivated = true;     % Put object to sleep if not moving
        
        for i = 2:50                             % Stairs:
            body(i).Geometry = 'cuboid';         % Geometry: 'cuboid' or 'sphere'
            body(i).HalfExtent = [ 10, (i-1)*0.1, 0.247]; % Dimensions, m
            body(i).M = 1e9;                     % Mass, kg
            body(i).X = [ 0, (i-1)*0.1, 15-i*0.5]; % Position, m
            body(i).Q = zeros(1,4);              % Orientation
            body(i).V = zeros(1,3);              % Velocity, m/s
            body(i).W = zeros(1,3);              % Angular velocity, rad/s
            body(i).ShowTrajectory = false;      % Don't track trajectories for stairs
            body(i).CanBeDeactivated = true;     % Can be deactived when not moving
            body(i).IsActive = false;            % Set object not moving
        end
        
        i = 51;
        body(i).Geometry = 'cuboid';         % Geometry: 'cuboid' or 'sphere'
        body(i).HalfExtent = [ 0.1, 8, 10 ]; % Dimensions, m
        body(i).M = 1e9;                     % Mass, kg
        body(i).X = [ -10.11, 8, 0 ]; % Position, m
        body(i).Q = zeros(1,4);              % Orientation
        body(i).V = zeros(1,3);              % Velocity, m/s
        body(i).W = zeros(1,3);              % Angular velocity, rad/s
        body(i).ShowTrajectory = false;      % Don't track trajectories for stairs
        body(i).CanBeDeactivated = true;     % Can be deactived when not moving
        body(i).IsActive = false;            % Set object not moving
        
        i = 52;
        body(i).Geometry = 'cuboid';         % Geometry: 'cuboid' or 'sphere'
        body(i).HalfExtent = [ 10, 8, 0.1 ]; % Dimensions, m
        body(i).M = 1e9;                     % Mass, kg
        body(i).X = [ 0, 8, -10.1 ]; % Position, m
        body(i).Q = zeros(1,4);              % Orientation
        body(i).V = zeros(1,3);              % Velocity, m/s
        body(i).W = zeros(1,3);              % Angular velocity, rad/s
        body(i).ShowTrajectory = false;      % Don't track trajectories for stairs
        body(i).CanBeDeactivated = true;     % Can be deactived when not moving
        body(i).IsActive = false;            % Set object not moving
        
        for i = 53:150
            body(i).Geometry = 'cuboid';         % Geometry: 'cuboid' or 'sphere'
            body(i).HalfExtent = 0.15 + ([0.6,0.4,0.3]-0.15).*rand(1,3); % Dimensions, m
            body(i).M = 10;                      % Mass, kg
            body(i).X = [ (i-100)/10*rand - 5, 12 + 3*rand, -8 ]; % Position, m
            body(i).Q = rand(1,4);               % Orientation
            body(i).V = zeros(1,3);              % Velocity, m/s
            body(i).W = rand(1,3);               % Angular velocity, rad/s
            body(i).CanBeDeactivated = true;     % Put object to sleep if not moving
            body(i).ActiveColor = [ rand(1,3), 0.8 ]; % Random object colors
        end
        
    case { '5a', '5b', '5b2', '5b3', '5c', '5d', '5e', '5f', '5g', '5h', '5i', '5j', '5k' }
        %% Pilling cuboid
            
        switch configId
            case '5a'
                nobj = 100;
                h = 0.01;
                M = 100;
                params.Restitution = 0.8;
                params.Relaxation = 0.2;
                params.CameraZoom = 25;
                params.CameraElevation = 55;
            case '5b' % N = 50
                nobj = 50;
                h = 0.01;
                M = 100;
                params.Restitution = 0.8;
                params.Relaxation = 0.2;
                params.CameraZoom = 25;
                params.CameraElevation = 50;
            case '5b2' % N = 20
                nobj = 25;
                h = 0.01;
                M = 100;
                params.Restitution = 0.8;
                params.Relaxation = 0.2;
                params.CameraZoom = 25;
                params.CameraElevation = 50;
            case '5b3' % N = 75
                nobj = 75;
                h = 0.01;
                M = 100;
                params.Restitution = 0.8;
                params.Relaxation = 0.2;
                params.CameraZoom = 25;
                params.CameraElevation = 50;
            case '5c' % N = 200
                nobj = 200;
                h = 0.01;
                M = 100;
                params.Restitution = 0.8;
                params.Relaxation = 0.2;
                params.CameraZoom = 35;
                params.CameraElevation = 70;
            case '5d' % COR = 0.7
                nobj = 100;
                h = 0.01;
                M = 100;
                params.Restitution = 0.7;
                params.Relaxation = 0.2;
                params.CameraZoom = 25;
                params.CameraElevation = 55;
            case '5e' % COR = 0.99
                nobj = 100;
                h = 0.01;
                M = 100;
                params.Restitution = 0.99;
                params.Relaxation = 0.2;
                params.CameraZoom = 25;
                params.CameraElevation = 55;
            case '5f' % kp = 0.1
                nobj = 100;
                h = 0.01;
                M = 100;
                params.Restitution = 0.8;
                params.Relaxation = 0.1;
                params.CameraZoom = 25;
                params.CameraElevation = 55;
            case '5g' % kp = 0.3
                nobj = 100;
                h = 0.01;
                M = 100;
                params.Restitution = 0.8;
                params.Relaxation = 0.3;
                params.CameraZoom = 25;
                params.CameraElevation = 55;
            case '5h' % h = 0.02
                nobj = 100;
                h = 0.02;
                M = 100;
                params.Restitution = 0.8;
                params.Relaxation = 0.2;
                params.CameraZoom = 25;
                params.CameraElevation = 55;
            case '5i' % h = 0.001
                nobj = 100;
                h = 0.001;
                M = 100;
                params.Restitution = 0.8;
                params.Relaxation = 0.2;
                params.CameraZoom = 25;
                params.CameraElevation = 55;
            case '5j'
                nobj = 100;
                h = 0.01;
                M = 50;
                params.Restitution = 0.8;
                params.Relaxation = 0.2;
                params.CameraZoom = 25;
                params.CameraElevation = 55;
            case '5k'
                nobj = 100;
                h = 0.01;
                M = 200;
                params.Restitution = 0.8;
                params.Relaxation = 0.2;
                params.CameraZoom = 25;
                params.CameraElevation = 55;
        end
        
        params.TimeStep = h;
        params.TimeStepsPerFrame = 0.01 / h;
        params.TimeStepsPerSnapshot = 50 * 0.01 / h;
        params.Friction = 0; 
    
        if params.TimeStepsPerFrame < 1
            params.TimeStepsPerFrame = 1;
        end
        
        params.IsPaused = true;
        params.FinalTime = 15;
        params.ShowTrajectories = true;
        params.ShowBodyAxes = false;
        params.PlotContacts = false;
        params.PlotEnergy = false;
        params.PlotLinearMomentum = false;
        params.PlotAngularMomentum = false;
        params.PlotHeight = true;
        params.FollowObject = -1;
        params.Gravity = [ 0, -9.80665, 0 ];
        params.CameraLookAt = [ 0, nobj * 0.2, 0 ];
        params.ShowHelp = false;
        
        for i = 1:nobj
            body(i).Geometry = 'cuboid';        % Geometry: 'cuboid' or 'sphere'
            body(i).HalfExtent = [ 1, 0.2, 1 ]; % Dimensions, m
            body(i).M = M;                      % Mass, kg
            body(i).X = [ 0, (i-1)*0.4, 0 ];    % Position, m
            body(i).Q = [ 1, 0, 0, 0 ];         % Orientation
            body(i).V = zeros(1,3);             % Velocity, m/s
            body(i).W = zeros(1,3);             % Angular velocity, rad/s
        end
        
    otherwise
        error( 'WoRB:confid', 'Unercognized configuration ID %s', configId );
    end
            

end % function